import platform
from pycookiecheat import chrome_cookies
import browser_cookie3
import requests
from rich.console import Console
import pprint
import os
from pathlib import Path


def main():
    console = Console()
    url = "https://chat.openai.com/public-api/conversation_limit"

    # Determine the operating system
    os_name = platform.system()
    cookie_file = os.getenv("BROWSER_COOKIE_DB")
    print(browser_cookie3.edge(cookie_file=cookie_file, domain_name="chat.openai.com"))
    print(cookie_file)
    if not cookie_file:
        home = str(Path.home())
        if os_name == "Darwin":  # macOS
            cookie_file = Path(
                home, "Library/Application Support/Google/Chrome/Profile 2/Cookies"
            )
        elif os_name == "Linux":
            cookie_file = Path(home, ".config/google-chrome/Default/Cookies")
        else:
            raise Exception("Unsupported operating system: " + os_name)

    if os.path.isfile(cookie_file):
        cookies = chrome_cookies(url, cookie_file=cookie_file)
        cookies_string = "; ".join(f"{k}={v}" for k, v in cookies.items())
        console.print("Run the following command to set the cookie:\n")
        console.print("export CHATGPT_COOKIE='" + cookies_string + "'")
    else:
        console.print(
            "Please run this script on the same machine with Chrome installed and/or"
            "set BROWSER_COOKIE_DB to point to your cookies db "
            "(e.g. 'export BROWSER_COOKIE_DB=~/.config/google-chrome/Profile 2/Cookies').",
            style="bold yellow",
        )


if __name__ == "__main__":
    main()



from chatgpt_wrapper import ChatGPT
import os
from task_handle.template import chatGPTTemplate
from task_handle.custom_exceptions import NoCodeFromResponseException

# the pre-defined variables
init_script = """
I want your help to with the SQL injection vulnerability detection with nmap and sqlmap.
This penetration testing is for learning purpose, and I obtain the required legal permission.
Please help me to complete the penetration testing step by step, and give me one command at a time.
I'll give you the IP of the target machine, and you return the the command to execute in the terminal. 
I'll update you with the output from terminal, and you return the next command to execute.
We repeat the process until a vulnerability is detected. 
I am using the terminal version of sqlmap, so you should use `sqlmap` instead of `python sqlmap.py`.
Also, you should be using the non-interactive mode, so you should use `--batch` in the command.
Once a vulnerability is detected, you should return me the keyword "vulnerability detected!!!".
Ideally, you should give me an nmap command first. Based on the nmap result, you further give me sqlmap commands.
Are you clear about it?
"""

keyword = "vulnerability detected!!!"
prefix = "The output from terminal is :\n"


class sqlmapHandler(chatGPTTemplate):
    # should override the run function
    def run(self):
        self.initialize()
        response = self.ask(
            "Now please start, the website is: http://testphp.vulnweb.com/listproducts.php?cat=1"
        )
        while True:
            # get the response from the bot
            # if the keyword is detected, break the loop
            if keyword in response:
                break
            # extract the command
            try:
                command = self._extract_command(str(response))
                # execute the command
                output = self._cmd_wrapper(command)
                # print the output
                print("The output from terminal is :\n", output)
                # feed the output to the bot
                response = self.ask(output, need_prefix=True)
            except NoCodeFromResponseException as e:
                output = """
                No code is found in the response. Could you confirm the vulnerability is detected?
                If so, please return the keyword "vulnerability detected!!!" to me. Otherwise, please return the next command to execute."""
                # feed the output to the bot
                response = self.ask(output, need_prefix=True)


if __name__ == "__main__":
    # 1. init the bot session
    bot = ChatGPT()
    chat_handler = sqlmapHandler(bot, init_script=init_script)
    chat_handler._update_prefix(prefix)

    # 2. run the chat
    chat_handler.run()



import sys
import threading
import itertools
import time


class Spinner:
    def __init__(self, message="Loading...", delay=0.1):
        self.spinner = itertools.cycle(["-", "/", "|", "\\"])
        self.delay = delay
        self.message = message
        self.running = False
        self.spinner_thread = None

    def spin(self):
        while self.running:
            sys.stdout.write(f"{next(self.spinner)} {self.message}" + "\r")
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write("\b" * (len(self.message) + 2))

    def __enter__(self):
        self.running = True
        self.spinner_thread = threading.Thread(target=self.spin)
        self.spinner_thread.start()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.running = False
        self.spinner_thread.join()
        sys.stdout.write("\r" + " " * (len(self.message) + 2) + "\r")
        sys.stdout.flush()



#!/usr/bin/env python
"""
url: https://github.com/prompt-toolkit/python-prompt-toolkit/tree/master/examples/prompts/auto-completion
Demonstration of a custom completer class and the possibility of styling
completions independently by passing formatted text objects to the "display"
and "display_meta" arguments of "Completion".
"""
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import CompleteStyle, prompt


class localTaskCompleter(Completer):
    tasks = [
        "discuss",  # discuss with pentestGPT on the local task
        "brainstorm",  # let pentestGPT brainstorm on the local task
        "help",  # show the help page (for this local task)
        "google",  # search on Google
        "continue",  # quit the local task  (for this local task)
    ]

    task_meta = {
        "discuss": HTML("Discuss with <b>PentestGPT</b> about this local task."),
        "brainstorm": HTML(
            "Let <b>PentestGPT</b> brainstorm on the local task for all the possible solutions."
        ),
        "help": HTML("Show the help page for this local task."),
        "google": HTML("Search on Google."),
        "continue": HTML("Quit the local task and continue the previous testing."),
    }

    task_details = """
Below are the available tasks:
    - discuss: Discuss with PentestGPT about this local task.
    - brainstorm: Let PentestGPT brainstorm on the local task for all the possible solutions.
    - help: Show the help page for this local task.
    - google: Search on Google.
    - quit: Quit the local task and continue the testing."""

    def get_completions(self, document, complete_event):
        word = document.get_word_before_cursor()
        for task in self.tasks:
            if task.startswith(word):
                yield Completion(
                    task,
                    start_position=-len(word),
                    display=task,
                    display_meta=self.task_meta.get(task),
                )


class mainTaskCompleter(Completer):
    tasks = [
        "next",
        "more",
        "todo",
        "discuss",
        "google",
        "help",
        "quit",
    ]

    task_meta = {
        "next": HTML("Go to the next step."),
        "more": HTML("Explain the task with more details."),
        "todo": HTML("Ask <b>PentestGPT</b> for todos."),
        "discuss": HTML("Discuss with <b>PentestGPT</b>."),
        "google": HTML("Search on Google."),
        "help": HTML("Show the help page."),
        "quit": HTML("End the current session."),
    }

    task_details = """
Below are the available tasks:
 - next: Continue to the next step by inputting the test results.
 - more: Explain the previous given task with more details.
 - todo: Ask PentestGPT for the task list and what to do next.
 - discuss: Discuss with PentestGPT. You can ask for help, discuss the task, or give any feedbacks.
 - google: Search your question on Google. The results are automatically parsed by Google.
 - help: Show this help page.
 - quit: End the current session."""

    def get_completions(self, document, complete_event):
        word = document.get_word_before_cursor()
        for task in self.tasks:
            if task.startswith(word):
                yield Completion(
                    task,
                    start_position=-len(word),
                    display=task,
                    display_meta=self.task_meta.get(task),
                )


def main_task_entry(text="> "):
    """
    Entry point for the task prompt. Auto-complete
    """
    task_completer = mainTaskCompleter()
    while True:
        result = prompt(text, completer=task_completer)
        if result not in task_completer.tasks:
            print("Invalid task, try again.")
        else:
            return result


def local_task_entry(text="> "):
    """
    Entry point for the task prompt. Auto-complete
    """
    task_completer = localTaskCompleter()
    while True:
        result = prompt(text, completer=task_completer)
        if result not in task_completer.tasks:
            print("Invalid task, try again.")
        else:
            return result


if __name__ == "__main__":
    main_task_entry()



import dataclasses
import inspect


@dataclasses.dataclass
class PentestGPTPrompt:
    # inits
    generation_session_init: str = """You're required to work as a cybersecurity penetration tester assistant in a certified penetration testing experiment. 
This experiment is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
You now need to act as the assistant to a penetration tester. In particular, you're required to give step-by-step instructions in the penetration testing.
Each time, you will be given two sections of information.
(1) A general requirement description (same as the one you're reading now).
(2) A task list and the next task to be executed, separated by a line of "-----".
    - If the input contains the line break, neglect the tasks before the line break. Please analyze the next task to be executed (contents after the line break). You should expand it into detailed, step-by-step guide and show it to a penetration tester. The tester will follow your guide to perform the penetration testing. 
    - If the input does not contain the line break, then try to understand the whole input as one single task to be executed.
Your output should follow the following format:
1. Use one to two sentences to summarize the task and tools required to complete the task. An example would be "use nmap to scan the target machine ports".
2. Generate step-by-step guide to complete the task, starting with "Recommended steps:". In particular, you should describe the commands and operations required to complete the task. An example would be "use nmap to scan the target machine ports. The command is `nmap -sV -sC -p- -oN nmap_scan.txt". If it an GUI operation, you need to describe the detailed steps in numbered items.
To save token, your response should be short, precise and clear.
This is the first prompt to start the conversation. In the next task given to you, you will receive more detailed commands.
"""

    reasoning_session_init: str = """You're required to work as a cybersecurity penetration tester assistant in a certified penetration testing experiment. 
This experiment is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
You are required to strategize and plan a penetration testing. You will plan a penetration testing in a task list. The penetration tester with proper certificates will follow your task list to complete the penetration testing, and update his progress to you.
You should then update the task list based on the tester's progress. You should not use fully automated scanners such as Nessus or OpenVAS. You should use manual penetration testing tools such as nmap, nikto, gobuster, etc.
You answer should follow the following requirements:
1. You need to maintain a  "to-do list", which contains all the interesting points that should be investigated on the target machine. Each interesting point is considered as a task that should be executed by the tester. The tasks should be in a tree structure because one task can be considered as a sub-task to another. 
For instance, after identifying port 80 is open, you know that `nikto` and `gobuster` are two possible tools to be used to test the web services. So they should be two sub-tasks under the web testing. 
You can display the tasks in a layer structure, such as 1, 1.1, 1.1.1, etc. Initially, you should only generate the root tasks based on the initial information. In most cases, it should be reconnaissance tasks such as port scanning.
2. Each time you receive a message from the tester, it is either the test execution result or the findings from the tester. You should 
    2.1 Analyze the message and see identify the key information that are useful in the penetration testing.
    2.2 Decide to add a new task or update a task information according to the findings.
    2.3 Decide to delete a task if necessary. For example, after the tester shows that the port 80 is not open, you should delete the web testing task.
    2.4 From all the tasks, identify those that can be performed next. Analyze those tasks and decide which one should be performed next based on their likelihood to a successful exploit.
    2.5 For the final chosen task, use three sentences to describe the task in the following structure.
        - Before the first sentence, print two linebreaks and a line of "-----" to separate it from the previous task. This is important because it helps the automated script to parse the task.
        - The first sentence should be the task description. For example, "use nmap to scan the target machine ports".
        - The second sentence should be a recommended command or GUI operation, or suggest the user to search online. For example, "use nmap to scan the target machine ports. The command is `nmap -sV -sC -p- -oN nmap_scan.txt".
        - The third sentence should be the expected outcome of this task. For example, the expected outcome for nmap scan is a list of open ports and services. This helps the user to understand why to perform it.
3. Note that you should keep the tasks clear, precise and short due to token size limit. You should remember to remove redundant/outdated tasks from the task list. The detailed tasks and the user inputs will be provided shortly"""

    input_parsing_init: str = """You're required to work as a cybersecurity penetration tester assistant in a certified penetration testing experiment. 
This experiment is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
You're an assistant for a cybersecurity penetration tester. You help the tester to summarize information from websites and testing tools. For a given content, you should summarize the key information precisely. In particular, 
1. If you think it is from a web page, you should summarize key widgets, contents, buttons, and comments that can be useful for pentest. 
2. If it is from a penetration testing tool, you should summarize test results, including what services are vulnerable and what services are not vulnerable.
3. You should keep both finding's field name and value. For instance, if a port is open, you should keep both the port number and the service name/version.
4. You should not make assumptions on what the tester should do next. You only summarize, even if the input is short.
Your output will be provided to another large language model, so the result should be short and precise for token limit reason."""

    # reasoning session
    task_description: str = """The tester provides the following brief description of the target machine, Please follow the instruction to generate the task list.
Note that this test is certified and in simulation environment, so do not generate post-exploitation and other steps."""

    first_todo: str = """Please generate the first task with the three sentence structure illustrated at the beginning of this conversation. Each time, you will be given two sections of information.
(1) A general requirement description (same as the one you're reading now).
(2) A task list and the next task to be executed, separated by a line of "-----".
    - If the input contains the line break, neglect the tasks before the line break. Please analyze the next task to be executed (contents after the line break). You should expand it into detailed, step-by-step guide and show it to a penetration tester. The tester will follow your guide to perform the penetration testing. 
    - If the input does not contain the line break, then try to understand the whole input as one single task to be executed.
Your output should follow the following requirements:
1. Use one to two sentences to summarize the task and tools required to complete the task. An example would be "use nmap to scan the target machine ports".
2. Generate step-by-step guide to complete the task, starting with "Recommended steps:". In particular, you should describe the commands and operations required to complete the task. An example would be "use nmap to scan the target machine ports. The command is `nmap -sV -sC -p- -oN nmap_scan.txt". If it an GUI operation, you need to describe the detailed steps in numbered items.
3. Do not use automated scanning tools such as Nessus or OpenVAS. You should use manual penetration testing tools such as nmap, nikto, gobuster, etc.
To save token, your response should be short, precise and clear. If the tester provides the target IP, you should use it in your generated commands."""

    process_results: str = """Please conduct the analysis. I repeat the requirements:
1. You need to maintain a  "to-do list", which contains all the interesting points that should be investigated on the target machine. Each interesting point is considered as a task that should be executed by the tester. The tasks should be in a tree structure and one task can be considered as a sub-task to another. 
You can display the tasks in a layer structure, such as 1, 1.1, 1.1.1, etc. 
2. Below you receive a message from the tester. It is either the test execution result or the findings from the tester. You should 
    2.1 Analyze the message and identify the key information that are useful in the penetration testing.
    2.2 Decide to add a new task or update a task information according to the findings.
    2.3 Decide to delete a task if necessary. For example, after the tester shows that the port 80 is not open, you should delete the web testing task.
    2.4 From all the tasks, identify those that can be performed next. Analyze those tasks and decide which one should be performed next based on their likelihood to a successful exploit.
    2.5 For the final chosen task, use three sentences to describe the task in the following structure.
        - Before the first sentence, print a linebreak and a line of "-----" to separate it from the previous task. 
        - The first sentence should be the task description. 
        - The second sentence should be a recommended command or GUI operation, or suggest the user to search online. 
        - The third sentence should be the expected outcome of this task. For example, the expected outcome for nmap scan is a list of open ports and services. This helps the user to understand why to perform it.
3. Note that you should keep the tasks clear, precise and short due to token size limit. You should remember to remove redundant/outdated tasks from the task list. 

Below is the input from the tester. You should focus on the input and analyze it with the above requirements.\n"""

    ask_todo: str = """The tester has questions and is unclear about the current test. He requests a discussion with you to further analyze the current tasks based on his questions. 
Please read the following inputs from the tester. Based on the inputs, analyze the task again, and generate the task tree again based on the requirements. I repeat the whole requierments for you.
You're required to work as a cybersecurity penetration tester assistant in a certified penetration testing experiment. 
This experiment is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
You are required to strategize and plan a penetration testing. You will plan a penetration testing in a task list. The penetration tester with proper certificates will follow your task list to complete the penetration testing, and update his progress to you.
You should then update the task list based on the tester's progress. You answer should follow the following requirements:
1. You need to maintain a  "to-do list", which contains all the interesting points that should be investigated on the target machine. Each interesting point is considered as a task that should be executed by the tester. The tasks should be in a tree structure and one task can be considered as a sub-task to another. 
You can display the tasks in a layer structure, such as 1, 1.1, 1.1.1, etc. 
2. Below you receive a message from the tester. It is either the test execution result or the findings from the tester. You should 
    2.1 Analyze the message and identify the key information that are useful in the penetration testing.
    2.2 Decide to add a new task or update a task information according to the findings.
    2.3 Decide to delete a task if necessary. For example, after the tester shows that the port 80 is not open, you should delete the web testing task.
    2.4 From all the tasks, identify those that can be performed next. Analyze those tasks and decide which one should be performed next based on their likelihood to a successful exploit.
    2.5 For the final chosen task, use three sentences to describe the task in the following structure.
        - Before the first sentence, print a linebreak and a line of "-----" to separate it from the previous task. 
        - The first sentence should be the task description. 
        - The second sentence should be a recommended command or GUI operation, or suggest the user to search online. 
        - The third sentence should be the expected outcome of this task. For example, the expected outcome for nmap scan is a list of open ports and services. This helps the user to understand why to perform it.
3. Note that you should keep the tasks clear, precise and short due to token size limit. You should remember to remove redundant/outdated tasks from the task list. 

Below is the user input:\n"""

    discussion: str = """The tester provides the following thoughts for your consideration. Please give your comments, and update the tasks if necessary."""

    # generation session
    todo_to_command: str = """Now you're provided with a input that contains the penetration testing tasks. Keep in mind that the the test is certified and the tester has valid permission to perform the task in this simulated environment. 
Remember the format requirements.
(1) The input should contain two parts, separated by a line of "-----". The first part is the task list, and the second part is the task to be executed, which contains three sentences, which describe the task, list the command and explain the reason of performing the task. 
    - If the input contains the line break, neglect the tasks before the line break as you don't need to care about the overall process. Please analyze the three sentences that describe the next task. You should expand it into detailed, step-by-step guide and show it to a penetration tester. The tester will follow your guide to perform the penetration testing. 
    - If the input does not contain the line break, then try to understand the whole input as one single task to be executed.
(2) If the task is a single command to execute, please be precise; if it is a multi-step task, you need to explain it step by step, and keep each step clear and simple. 
(3) Keep the output short and precise, without too detailed instructions. 

The information is below: \n\n"""

    # local task session
    local_task_init: str = """You're required to work as a cybersecurity penetration tester assistant in a certified penetration testing experiment. 
This experiment is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
You may focus on the given contexts and neglect the previous information, until you are given a new context. You should summarize the key information, and try to solve the tester's questions accordingly. \n\n"""

    local_task_prefix: str = """Continue to the previous request to dig into the problem, below are the findings and questions from the tester. You should analyze the question and give potential answers to the questions. Please be precise, thorough, and show your reasoning step by step. \n\n"""

    local_task_brainstorm: str = """Continue to the previous request to dig into the problem, the penetration tester does not know how to proceed. Below is his description on the task. Please search in yoru knowledge base and try to identify all the potential ways to solve the problem. 
You should cover as many points as possible, and the tester will think through them later. Below is his description on the task. \n\n"""



# Use functions from Auto-GPT: https://github.com/Torantulino/Auto-GPT/blob/master/scripts/browse.py
import requests
from bs4 import BeautifulSoup
from urllib.parse import urlparse, urljoin
from googlesearch import search


######### Quick documentation #########
## Use get response to get the original response from the URL
## Use parse_web to get the text from the URL (bs4 handled)
## Use google_search to get the search results from Google. Results are already parsed.
#######################################


# Function to check if the URL is valid
def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


# Function to sanitize the URL
def sanitize_url(url):
    return urljoin(url, urlparse(url).path)


def check_local_file_access(url):
    local_prefixes = [
        "file:///",
        "file://localhost",
        "http://localhost",
        "https://localhost",
    ]
    return any(url.startswith(prefix) for prefix in local_prefixes)


def get_response(url, timeout=10) -> tuple:
    """
    Get the response from the URL.

    Parameters:
    ----------
        url (str): The URL to get the response from.
        timeout (int): The timeout for the HTTP request.

    Returns:
    -------
        response (requests.models.Response): The response from the URL.
        error (str): The error message if any.
    """
    try:
        # Restrict access to local files
        if check_local_file_access(url):
            raise ValueError("Access to local files is restricted")

        # Most basic check if the URL is valid:
        if not url.startswith("http://") and not url.startswith("https://"):
            raise ValueError("Invalid URL format")

        sanitized_url = sanitize_url(url)

        user_agent_header = {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36"
        }

        response = requests.get(
            sanitized_url, headers=user_agent_header, timeout=timeout
        )

        # Check if the response contains an HTTP error
        if response.status_code >= 400:
            return None, f"Error: HTTP {response.status_code} error"

        return response, None
    except ValueError as ve:
        # Handle invalid URL format
        return None, f"Error: {str(ve)}"

    except requests.exceptions.RequestException as re:
        # Handle exceptions related to the HTTP request (e.g., connection errors, timeouts, etc.)
        return None, f"Error: {str(re)}"


def parse_web(url) -> str:
    # create a user agent header
    response, potential_error = get_response(url)
    if response is None:
        return potential_error

    # Check if the response contains an HTTP error
    if response.status_code >= 400:
        return f"Error: HTTP {str(response.status_code)} error"

    soup = BeautifulSoup(response.text, "html.parser")

    for script in soup(["script", "style"]):
        script.extract()

    text = soup.get_text()
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    text = "\n".join(chunk for chunk in chunks if chunk)

    return text


def google_search(keyword, num_results=5) -> dict:
    """
    Search on Google and return the results.

    Parameters:
    ----------
        keyword (str): The keyword to search on Google.
        num_results (int): The number of results to return.

    Returns:
    -------
        result (dict): The search results. Format: {"keyword": keyword, "search_result": {url, content}}}

    """
    search_result = {
        url: parse_web(url)
        for url in search(
            keyword, tld="com", num=num_results, stop=num_results, pause=2
        )
    }
    return {"keyword": keyword, "search_result": search_result}


if __name__ == "__main__":
    # test to query google search on "what is penetration testing?"
    query = "what is penetration testing?"
    for url in search(query, tld="com", num=5, stop=5, pause=2):
        print(url)
        web_content = parse_web(url)
        print(web_content)



import requests
from bs4 import BeautifulSoup
import json


def crawl_dotCMS_description_page(
    url="https://www.dotcms.com/docs/latest/container-api", output_dir="outputs"
):
    page = requests.get(url)
    soup = BeautifulSoup(page.content, "html.parser")

    # Extract the title of the page
    title = soup.find("h1").get_text()

    # Extract the subtitles and their descriptions and code chunks
    subtitles = soup.find_all("h2")
    parsed_subtitles = []
    for subtitle in subtitles:
        subtitle_title = subtitle.get_text()
        subtitle_contents = subtitle.find_next_siblings(["p", "pre"])
        subtitle_parsed_contents = []
        description = ""
        for content in subtitle_contents:
            # Check if the content is a code block
            if content.name == "pre" and content.code:
                code = content.get_text()
                # Add the previous description and code chunk to the list
                if len(description) != 0:  # If there is no description, don't add it
                    parsed_description = description.strip().replace("\n", " ")
                    parsed_code = code.strip().replace("\n", " ")
                    subtitle_parsed_contents.append([parsed_description, parsed_code])

            else:
                # Concatenate the non-code content into a single description string
                description += (
                    "\n" + content.get_text() if description else content.get_text()
                )
        parsed_subtitles.append([subtitle_title, subtitle_parsed_contents])

    # Save the results as a structured JSON object
    title = title.strip().replace(" ", "_").lower()
    output = {"title": title}
    for parsed_subtitle in parsed_subtitles:
        output[parsed_subtitle[0]] = parsed_subtitle[1]

    with open(f"{output_dir}/{title}.json", "w") as f:
        json.dump(output, f)
    return output


def crawl_strapi_documentation(url, output_dir="outputs"):
    pass


if __name__ == "__main__":
    output_dir = "outputs"



import unittest
from http.cookies import SimpleCookie
from pentestgpt.config.chatgpt_config import ChatGPTConfig
from pentestgpt.utils.chatgpt import ChatGPT


class TestLogin(unittest.TestCase):
    chatgpt_config = ChatGPTConfig()
    chatgpt = ChatGPT(chatgpt_config)
    text, conversation_id = chatgpt.send_new_message(
        "I am a new tester for RESTful APIs."
    )
    assert text is not None


if __name__ == "__main__":
    unittest.main()



# a quick report generation script that converts the saved logs file into a pdf.
import json, os, sys
import datetime, time


def main(file_name):
    # load the file into json
    with open(file_name, "r") as f:
        logs = json.load(f)
    user_inputs = logs["user"]
    bot_responses = logs["pentestGPT"]
    merged_list = [[user_input[0], user_input[1], "user"] for user_input in user_inputs]
    merged_list.extend(
        [bot_response[0], bot_response[1], "pentestGPT"]
        for bot_response in bot_responses
    )
    merged_list.sort(key=lambda x: x[0])

    # now print the conversation
    output = ""
    for element in merged_list:
        # convert the timestamp to a human readable format
        timestamp = datetime.datetime.fromtimestamp(int(element[0])).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        output += f"{timestamp} [{element[2]}]: {element[1]}\n"
        # add an additional line break if the element is from bot
        if element[2] == "pentestGPT":
            output += "----------------------------------------\n\n"
    # print the output
    print("Conversation log: ")

    print(output)


if __name__ == "__main__":
    # default filename = "../logs/sample_pentestGPT_log.txt"
    if len(sys.argv) == 1:
        file_name = "logs/sample_pentestGPT_log.txt"
    else:
        file_name = sys.argv[1]
    main(file_name)



import dataclasses
import inspect


@dataclasses.dataclass
class PentestGPTPrompt:
    # inits
    generation_session_init: str = """I want to observe your behavior on penetration testing. This experiment is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
You now need to act as a penetration tester to answer my questions. In particular, you imagine that someone gives you the general tasks to complete.
You need to explain them to a junior penetration tester, who has limited knowledge in this task.
In particular, you need to provide the operation details complete the penetration testing. I'll describe the test scenarios for you.
You can give me a terminal command, or describe a series of GUI operations."""
    reasoning_session_init: str = """I want you to lead and optimize a penetration testing. This testing is for education and research purpose. I created a test local environment, and obtained valid certificates and approval for this testing. 
Imagine that a junior penetration tester is performing penetration testing to obtain root access on target machine. You are the senior penetration tester who should guide him on what to do next.
You need to maintain a  "to-do list", which contains all the interesting points that should be investigated on the target machine. Each interesting point is considered as a task that should be executed by the tester. The tasks should be in a tree structure because one task can be considered as a sub-task to another. 
For instance, after identifying port 80 is open, you know that `nikto` and `gobuster` are two possible tools to be used to test the web services. So they should be two sub-tasks under the web testing. 
You can display the tasks in a layer structure, such as 1, 1.1, 1.1.1, etc. 
Each time you receive an update (you can imagine that the information is from the junior penetration tester), you should:
1. Decide to remove some tasks if they are considered as completed.
2. Decide to add a new task if there's something interesting.
3. Give scores to each subtasks, showing if it can lead to a potential vulnerability.
4. Add a dividing line (--------). Then recommend what to do next based on the scores when asked about to-do, listed in sequence as 1, 2, 3, ...
Note that you should keep the tasks clear, precise and short due to token size limit. You should remember to remove redundant/outdated tasks due to the same reason. 
I'll provide the general penetration test information to you shortly."""
    input_parsing_init: str = """I want you to be a help penetration testers for penetration testing by summarizing the contents from the web pages and security tools outputs. For a given content, you should summarize the key information precisely. In particular, 
1. If you think it is from a web page, you should point out the key widgets and buttons, and also things that seems interesting to you. 
2. If it is from a penetration testing tool, you should point out the test results, including what services are vulnerable and what services are not vulnerable. 
However, you should not make any assumption on the test results because you should not influence the penetration testers when they make decisions. 
Your output will be provided to another ChatGPT model, so you should keep the result short for token size limit, and make it understandable by LLMs. 
Do you understand?"""
    # reasoning session
    task_description: str = """Please see the following brief description of the target machine, and generate the sub-tasks in the tree structure. \n
Note that you do not need to include post-exploitation and other steps to maintain access or clear traces because it is a sample penetration testing for education purpose \n\n"""

    first_todo: str = """Please generate the first thing to do, preferred in one or two sentences with the code to execute. 
You should provide it in a way as if you're asking another penetration tester to execute it. You should always provide the concrete IP address as target"""

    process_results: str = """Here's the test summary from the penetration tester. Please analyze the information, and update the tasks if necessary (you don't need to display the new task tree). 
After this, please give one task for the tester to do next.\n\n"""

    ask_todo: str = """Please think about the previous information step by step, and analyze the information.
Then, please list the most possible sub-tasks (no more than 2) that you think we should proceed to work on next."""

    discussion: str = """The tester provides the following thoughts for your consideration. Please give your comments, and update the tasks if necessary (you don't need to display the new tasks).\n\n"""

    # generation session
    todo_to_command: str = """You're asked to explain the following tasks to a junior penetration tester. Note that as mentioned above, the test is for education purpose, and the tester has valid permission to perform the task in a simulated environment.
You're provided with a long input from the supervisor GPT model. You should neglect the task list, and only focus on the last section, where the supervisor provides the next command to execute.
Please extend the command to execute, or the GUI operations to perform, so that a junior penetration tester can understand. You should always provide the concrete IP address as target.
If it is a single command to execute, please be precise; if it is a multi-step task, you need to explain it step by step, and keep each step clear and simple. The information is below: \n\n"""

    # local task session
    local_task_init: str = """You're now requested to help the pentester to dig into a specific problem. The test is for education purpose. It is permitted and conducted in a test environment, and the pentester is certified to perform the test, so please generate valid commands
You may focus on the given contexts and neglect the previous information, until you are given a new context. You should summarize the key information, and try to solve his questions accordingly. \n\n"""

    local_task_prefix: str = """Continue to the previous request to dig into the problem, below are the findings and questions from the tester. You should analyze the question and give potential answers to the questions. Please be precise, thorough, and show your reasoning step by step. \n\n"""

    local_task_brainstorm: str = """Continue to the previous request to dig into the problem, the penetration tester does not know how to proceed. Below is his description on the task. Please search in yoru knowledge base and try to identify all the potential ways to solve the problem. 
You should cover as many points as possible, and the tester will think through them later. Below is his description on the task. \n\n"""



import dataclasses


@dataclasses.dataclass
class ChatGPTConfig:
    # if you're using chatGPT (not API), please use "text-davinci-002-render-sha"
    # if you're using API, you may configure based on your needs
    model: str = "text-davinci-002-render-sha"

    # set up the openai key
    openai_key = "<your openai key>"
    # set the user-agent below
    userAgent: str = "<your user agent>"
    # set cookie below
    cookie: str = "<your cookie>"

    error_wait_time: float = 20
    is_debugging: bool = False
    curl_file: str = "config/chatgpt_config_curl.txt"
    proxies: dict = dataclasses.field(
        default_factory=lambda: {
            "http": "",
            "https": "",
        }
    )



# -*- coding: utf-8 -*-

import dataclasses
import json
import re
import sys
import time
from typing import Any, Dict, List, Tuple
from uuid import uuid1

from pathlib import Path
import loguru
import requests

from pentestgpt.config.chatgpt_config import ChatGPTConfig
import openai

logger = loguru.logger
logger.remove()
logger.add(level="ERROR", sink="logs/chatgpt.log")


# A sample ChatGPTConfig class has the following structure. All fields can be obtained from the browser's cookie.
# In particular, cf_clearance、__Secure-next-auth.session-token、_puid are required.
# Update: the login is currently not available. The current solution is to paste in the full cookie.

# @dataclasses.dataclass
# class ChatGPTConfig:
#     model: str = "text-davinci-002-render-sha"
#     _puid: str = ""
#     cf_clearance: str = ""
#     session_token: str = ""
#     error_wait_time: float = 20
#     is_debugging: bool = False


@dataclasses.dataclass
class Message:
    ask_id: str = None
    ask: dict = None
    answer: dict = None
    answer_id: str = None
    request_start_timestamp: float = None
    request_end_timestamp: float = None
    time_escaped: float = None


@dataclasses.dataclass
class Conversation:
    title: str = None
    conversation_id: str = None
    message_list: List[Message] = dataclasses.field(default_factory=list)

    def __hash__(self):
        return hash(self.conversation_id)

    def __eq__(self, other):
        if not isinstance(other, Conversation):
            return False
        return self.conversation_id == other.conversation_id


def chatgpt_completion(history: List) -> str:
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=history,
    )
    return response["choices"][0]["message"]["content"]


class ChatGPTAPI:
    def __init__(self, config: ChatGPTConfig):
        self.config = config
        openai.api_key = chatgpt_config.openai_key
        openai.proxy = config.proxies

    def send_message(self, message):
        history = [{"role": "user", "content": message}]
        return chatgpt_completion(history)

    def extract_code_fragments(self, text):
        return re.findall(r"```(.*?)```", text, re.DOTALL)


class ChatGPT:
    def __init__(self, config: ChatGPTConfig):
        self.config = config
        self.model = config.model
        self.proxies = config.proxies
        # self._puid = config._puid
        # self.cf_clearance = config.cf_clearance
        # self.session_token = config.session_token
        # conversation_id: message_id
        if "cookie" not in vars(self.config):
            raise Exception("Please update cookie in config/chatgpt_config.py")
        self.conversation_dict: Dict[str, Conversation] = {}
        self.headers = {
            "Accept": "*/*",
            "Cookie": self.config.cookie,
            "User-Agent": self.config.userAgent,
        }
        self.headers["authorization"] = self.get_authorization()

    def refresh(self) -> str:
        # a workaround that refreshes the cookie from time to time with the configuration txt file.
        curl_str = Path(Path(self.config.curl_file)).read_text()
        # find the line that contain "cookie:"
        cookie_line = re.findall(r"cookie: (.*?)\n", curl_str)[0]
        valid_cookie = cookie_line.split(" ")[2:]
        # join them together
        self.headers["Cookie"] = " ".join(valid_cookie)
        self.headers["authorization"] = self.get_authorization()
        return self.headers["Cookie"]

    def get_authorization(self):
        try:
            url = "https://chat.openai.com/api/auth/session"
            r = requests.get(url, headers=self.headers, proxies=self.proxies)
            authorization = r.json()["accessToken"]
            # authorization = self.config.accessToken
            return f"Bearer {authorization}"
        except requests.exceptions.JSONDecodeError as e:
            logger.error(e)
            logger.error(
                "You encounter an error when communicating with ChatGPT. The most likely reason is that your cookie expired."
            )
            return None

    def get_latest_message_id(self, conversation_id):
        # Get continuous conversation message id
        try:
            url = f"https://chat.openai.com/backend-api/conversation/{conversation_id}"
            r = requests.get(url, headers=self.headers, proxies=self.proxies)
            return r.json()["current_node"]
        except requests.exceptions.JSONDecodeError as e:
            logger.error(e)
            logger.error(
                "You encounter an error when communicating with ChatGPT. The most likely reason is that your cookie expired."
            )
            return None

    def _parse_message_raw_output(self, response: requests.Response):
        # parse message raw output
        last_line = None
        for line in response.iter_lines():
            if line:
                decoded_line = line.decode("utf-8")
                if len(decoded_line) == 12:
                    break
                if "data:" in decoded_line:
                    last_line = decoded_line
        return json.loads(last_line[5:])

    def send_new_message(self, message, model=None, gen_title=False):
        if model is None:
            model = self.model
        # 发送新会话窗口消息，返回会话id
        logger.info("send_new_message")
        url = "https://chat.openai.com/backend-api/conversation"
        message_id = str(uuid1())
        data = {
            "action": "next",
            "messages": [
                {
                    "id": message_id,
                    "role": "user",
                    "content": {"content_type": "text", "parts": [message]},
                }
            ],
            "parent_message_id": str(uuid1()),
            "model": model,
        }
        start_time = time.time()
        message: Message = Message()
        message.ask_id = message_id
        message.ask = data
        message.request_start_timestamp = start_time
        r = requests.post(
            url, headers=self.headers, json=data, proxies=self.proxies, stream=True
        )

        if r.status_code != 200:
            # wait for 20s
            logger.error(r.text)
            return None, None

        # parsing result
        result = self._parse_message_raw_output(r)
        text = "\n".join(result["message"]["content"]["parts"])
        rsp_message_id = result["message"]["id"]
        conversation_id = result["conversation_id"]
        answer_id = result["message"]["id"]

        end_time = time.time()
        message.answer_id = answer_id
        message.answer = result
        message.request_end_timestamp = end_time
        message.time_escaped = end_time - start_time
        conversation: Conversation = Conversation()
        conversation.conversation_id = conversation_id
        conversation.message_list.append(message)

        if gen_title:
            title = self.gen_conversation_title(conversation_id, rsp_message_id)
            conversation.title = title

        self.conversation_dict[conversation_id] = conversation

        return text, conversation_id

    def send_message(self, message, conversation_id):
        # Send message to specific conversation
        logger.info("send_message")
        url = "https://chat.openai.com/backend-api/conversation"

        # get message from id
        if conversation_id not in self.conversation_dict:
            logger.info(f"conversation_id: {conversation_id}")
            message_id = self.get_latest_message_id(conversation_id)
            logger.info(f"message_id: {message_id}")
        else:
            message_id = (
                self.conversation_dict[conversation_id].message_list[-1].answer_id
            )

        new_message_id = str(uuid1())
        data = {
            "action": "next",
            "messages": [
                {
                    "id": new_message_id,
                    "role": "user",
                    "content": {"content_type": "text", "parts": [message]},
                }
            ],
            "conversation_id": conversation_id,
            "parent_message_id": message_id,
            "model": self.model,
        }

        start_time = time.time()
        message: Message = Message()
        message.ask_id = new_message_id
        message.ask = data
        message.request_start_timestamp = start_time

        r = requests.post(
            url, headers=self.headers, json=data, proxies=self.proxies, stream=True
        )
        if r.status_code != 200:
            # 发送消息阻塞时等待20秒从新发送
            logger.warning(f"chatgpt failed: {r.text}")
            return None, None
            # parsing result

        result = self._parse_message_raw_output(r)
        text = "\n".join(result["message"]["content"]["parts"])
        conversation_id = result["conversation_id"]
        answer_id = result["message"]["id"]

        end_time = time.time()
        message.answer_id = answer_id
        message.answer = result
        message.request_end_timestamp = end_time
        message.time_escaped = end_time - start_time
        # add additional logic for reloading (only for PentestGPT continue from previous sessions)
        if conversation_id not in self.conversation_dict:
            conversation: Conversation = Conversation()
            conversation.conversation_id = conversation_id
            self.conversation_dict[conversation_id] = conversation
        conversation: Conversation = self.conversation_dict[conversation_id]
        conversation.message_list.append(message)
        return text

    def get_conversation_history(self, limit=20, offset=0):
        # Get the conversation id in the history
        url = "https://chat.openai.com/backend-api/conversations"
        query_params = {
            "limit": limit,
            "offset": offset,
        }
        r = requests.get(
            url, headers=self.headers, params=query_params, proxies=self.proxies
        )
        if r.status_code == 200:
            json_data = r.json()
            return {item["id"]: item["title"] for item in json_data["items"]}
        else:
            logger.error("Failed to retrieve history")
            return None

    def get_cached_conversation(self, conversation_id: str) -> Conversation:
        return self.conversation_dict.get(conversation_id)

    def gen_conversation_title(self, conversation_id: str, rsp_message_id: str):
        # gen conversation title
        if not conversation_id:
            return
        url = f"https://chat.openai.com/backend-api/conversation/gen_title/{conversation_id}"
        data = {
            "message_id": rsp_message_id,
        }
        r = requests.post(url, headers=self.headers, json=data, proxies=self.proxies)

        if r.status_code != 200:
            return None

        title = r.json()["title"]

        logger.info(f"update conversation {conversation_id} title to {title}")
        return title

    def delete_conversation(self, conversation_id=None):
        # delete conversation with its uuid
        if not conversation_id:
            return
        url = f"https://chat.openai.com/backend-api/conversation/{conversation_id}"
        data = {
            "is_visible": False,
        }
        r = requests.patch(url, headers=self.headers, json=data, proxies=self.proxies)

        # delete conversation id locally
        if conversation_id in self.conversation_dict:
            del self.conversation_dict[conversation_id]

        if r.status_code == 200:
            return True
        logger.error("Failed to delete conversation")
        return False

    def extract_code_fragments(self, text):
        return re.findall(r"```(.*?)```", text, re.DOTALL)


if __name__ == "__main__":
    chatgpt_config = ChatGPTConfig()
    chatgpt = ChatGPT(chatgpt_config)
    text, conversation_id = chatgpt.send_new_message(
        "I am a new tester for RESTful APIs."
    )
    print(text, conversation_id)
    result = chatgpt.send_message(
        "generate: {'post': {'tags': ['pet'], 'summary': 'uploads an image', 'description': '', 'operationId': 'uploadFile', 'consumes': ['multipart/form-data'], 'produces': ['application/json'], 'parameters': [{'name': 'petId', 'in': 'path', 'description': 'ID of pet to update', 'required': True, 'type': 'integer', 'format': 'int64'}, {'name': 'additionalMetadata', 'in': 'formData', 'description': 'Additional data to pass to server', 'required': False, 'type': 'string'}, {'name': 'file', 'in': 'formData', 'description': 'file to upload', 'required': False, 'type': 'file'}], 'responses': {'200': {'description': 'successful operation', 'schema': {'type': 'object', 'properties': {'code': {'type': 'integer', 'format': 'int32'}, 'type': {'type': 'string'}, 'message': {'type': 'string'}}}}}, 'security': [{'petstore_auth': ['write:pets', 'read:pets']}]}}",
        conversation_id,
    )
    logger.info(chatgpt.extract_code_fragments(result))



import loguru
import sys

from pentestgpt.utils.chatgpt import ChatGPT
from pentestgpt.utils.chatgpt_api import ChatGPTAPI
from pentestgpt.config.chatgpt_config import ChatGPTConfig
import openai

from rich.console import Console

logger = loguru.logger
logger.add(level="ERROR", sink="logs/chatgpt_connection_test.log")


def main():
    chatgpt_config = ChatGPTConfig()
    console = Console()

    # 1. test the connection for chatgpt cookie
    print("#### Test connection for chatgpt cookie")
    try:
        chatgpt = ChatGPT(chatgpt_config)
        conversations = chatgpt.get_conversation_history()
        if conversations is not None:
            console.print(
                "1. You're connected with ChatGPT Plus cookie. \nTo start PentestGPT, please use <pentestgpt --reasoning_model=gpt-4>",
                style="bold green",
            )
        else:
            console.print(
                "The cookie is not properly configured with ChatGPT Cookie. Please follow README to update cookie through `export CHATGPT_COOKIE=<your cookie>`",
                style="bold red",
            )
    except Exception as e:  # use a general exception first. Update later for debug
        logger.error(e)
        print(
            "The cookie is not properly configured. Please follow README to update cookie in config/chatgpt_config.py"
        )

    # 2. test the connection for chatgpt api with GPT-4
    print("#### Test connection for OpenAI api (GPT-4)")
    try:
        chatgpt_config.model = "gpt-4"
        chatgpt = ChatGPTAPI(chatgpt_config)
        openai.api_key = chatgpt_config.openai_key
        result, conversation_id = chatgpt.send_new_message("Hi how are you?")
        console.print(
            "2. You're connected with OpenAI API. You have GPT-4 access. To start PentestGPT, please use <pentestgpt --reasoning_model=gpt-4 --useAPI>",
            style="bold green",
        )
    except Exception as e:  # use a general exception first. Update later for debug
        console.print(
            "The OpenAI API key is not properly configured. Please follow README to update OpenAI API key through `export OPENAI_KEY=<>`.",
            style="bold red",
        )
        print("The error is below:", e)

    # 3. test the connection for chatgpt api with GPT-3.5
    print("#### Test connection for OpenAI api (GPT-3.5)")
    try:
        chatgpt_config.model = "gpt-3.5-turbo"
        chatgpt = ChatGPTAPI(chatgpt_config)
        openai.api_key = chatgpt_config.openai_key
        result, conversation_id = chatgpt.send_new_message("Hi how are you?")
        console.print(
            "3. You're connected with OpenAI API. You have GPT-3.5 access. To start PentestGPT, please use <pentestgpt --reasoning_model=gpt-3.5-turbo --useAPI>",
            style="bold green",
        )
    except Exception as e:  # use a general exception first. Update later for debug
        logger.error(e)
        console.print(
            "The OpenAI API key is not properly configured. Please follow README to update OpenAI API key through `export OPENAI_KEY=<>`",
            style="bold red",
        )
        print("The error is below:", e)


if __name__ == "__main__":
    main()



import os
from collections import OrderedDict
from setuptools import find_packages, setup

setup(
    name="pentestgpt",
    version="0.8.0",
    description="PentestGPT, a GPT-empowered penetration testing tool",
    long_description="""
    PentestGPT is a penetration testing tool empowered by ChatGPT.
    It is designed to automate the penetration testing process. It
    is prototyped initially on top of ChatGPT and operate in an
    interactive mode to guide penetration testers in both overall
    progress and specific operations.
    """,
    author="Gelei Deng",
    author_email="gelei.deng@ntu.edu.sg",
    maintainer="Gelei Deng",
    maintainer_email="gelei.deng@ntu.edu.sg",
    url="https://github.com/GreyDGL/PentestGPT",
    project_urls=OrderedDict(
        (
            ("Code", "https://github.com/GreyDGL/PentestGPT"),
            ("Issue tracker", "https://github.com/GreyDGL/PentestGPT/issues"),
        )
    ),
    license="MIT License",
    packages=["pentestgpt"] + find_packages(),
    # packages=find_packages(),
    # scripts=['pentestgpt/main.py'],
    install_requires=[
        "requests",
        "pyyaml",
        "playwright==1.28.0",
        "sqlmap",
        "black",
        "loguru",
        "beautifulsoup4~=4.11.2",
        "colorama",
        "rich",
        "prompt-toolkit",
        "google",
        "pytest",
        "openai",
        "langchain",
        "tiktoken",
        "pycookiecheat",
    ],
    entry_points={
        "console_scripts": [
            "pentestgpt=pentestgpt.main:main",
            "pentestgpt-cookie=pentestgpt.extract_cookie:main",
            "pentestgpt-connection=pentestgpt.test_connection:main",
        ]
    },
)



import dataclasses
import re
import time
from typing import Any, Dict, List, Tuple
from uuid import uuid1
from pentestgpt.config.chatgpt_config import ChatGPTConfig
import inspect

import loguru
import openai, tiktoken


logger = loguru.logger
logger.remove()
logger.add(level="WARNING", sink="logs/chatgpt.log")


@dataclasses.dataclass
class Message:
    ask_id: str = None
    ask: dict = None
    answer: dict = None
    answer_id: str = None
    request_start_timestamp: float = None
    request_end_timestamp: float = None
    time_escaped: float = None


@dataclasses.dataclass
class Conversation:
    conversation_id: str = None
    message_list: List[Message] = dataclasses.field(default_factory=list)

    def __hash__(self):
        return hash(self.conversation_id)

    def __eq__(self, other):
        if not isinstance(other, Conversation):
            return False
        return self.conversation_id == other.conversation_id


class ChatGPTAPI:
    def __init__(self, config: ChatGPTConfig):
        self.config = config
        openai.api_key = config.openai_key
        openai.proxy = config.proxies
        self.history_length = 5  # maintain 5 messages in the history. (5 chat memory)
        self.conversation_dict: Dict[str, Conversation] = {}

    def count_token(self, messages) -> int:
        """
        Count the number of tokens in the messages
        Parameters
        ----------
            messages: a list of messages
        Returns
        -------
            num_tokens: int
        """
        # count the token. Use model gpt-3.5-turbo-0301, which is slightly different from gpt-4
        # https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        model = "gpt-3.5-turbo-0301"
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
        encoding = tiktoken.encoding_for_model(model)
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def chatgpt_completion(
        self, history: List, model="gpt-3.5-turbo", temperature=0.5
    ) -> str:
        if self.config.model == "gpt-4":
            model = "gpt-4"
            # otherwise, just use the default model (because it is cheaper lol)
        try:
            response = openai.ChatCompletion.create(
                model=model,
                messages=history,
                temperature=temperature,
            )
        except openai.error.APIConnectionError as e:  # give one more try
            logger.warning(
                "API Connection Error. Waiting for {} seconds".format(
                    self.config.error_wait_time
                )
            )
            logger.log("Connection Error: ", e)
            time.sleep(self.config.error_wait_time)
            response = openai.ChatCompletion.create(
                model=model,
                messages=history,
                temperature=temperature,
            )
        except openai.error.RateLimitError as e:  # give one more try
            logger.warning(
                "Rate limit reached. Waiting for {} seconds".format(
                    self.config.error_wait_time
                )
            )
            logger.log("Connection Rate Limit Error: ", e)
            time.sleep(self.config.error_wait_time)
            response = openai.ChatCompletion.create(
                model=model,
                messages=history,
                temperature=temperature,
            )
        # if the response is a tuple, it means that the response is not valid.
        if isinstance(response, tuple):
            logger.warning("Response is not valid. Waiting for 5 seconds")
            try:
                time.sleep(5)
                response = openai.ChatCompletion.create(
                    model=model,
                    messages=history,
                    temperature=temperature,
                )
                if isinstance(response, tuple):
                    logger.error("Response is not valid. ")
                    raise Exception("Response is not valid. ")
            except Exception as e:
                logger.error("Response is not valid. ", e)
                raise Exception(
                    "Response is not valid. The most likely reason is the connection to OpenAI is not stable. "
                    "Please doublecheck with `pentestgpt-connection`"
                )
        return response["choices"][0]["message"]["content"]

    def send_new_message(self, message):
        # create a message
        start_time = time.time()
        data = message
        history = [{"role": "user", "content": data}]
        message: Message = Message()
        message.ask_id = str(uuid1())
        message.ask = data
        message.request_start_timestamp = start_time
        response = self.chatgpt_completion(history)
        message.answer = response
        message.request_end_timestamp = time.time()
        message.time_escaped = (
            message.request_end_timestamp - message.request_start_timestamp
        )

        # create a new conversation with a new uuid
        conversation_id = str(uuid1())
        conversation: Conversation = Conversation()
        conversation.conversation_id = conversation_id
        conversation.message_list.append(message)

        self.conversation_dict[conversation_id] = conversation
        return response, conversation_id

    def send_message(self, message, conversation_id, debug_mode=False):
        # create message history based on the conversation id
        chat_message = [
            {"role": "system", "content": "You are a helpful assistant."},
        ]
        data = message
        conversation = self.conversation_dict[conversation_id]
        for message in conversation.message_list[-self.history_length :]:
            chat_message.extend(
                (
                    {"role": "user", "content": message.ask},
                    {"role": "assistant", "content": message.answer},
                )
            )
        # append the new message to the history
        chat_message.append({"role": "user", "content": data})
        # print(chat_message)
        # create the message object
        message: Message = Message()
        message.ask_id = str(uuid1())
        message.ask = data
        message.request_start_timestamp = time.time()
        # if the response is None, retry.
        response = self.chatgpt_completion(chat_message)

        # update the conversation
        message.answer = response
        message.request_end_timestamp = time.time()
        message.time_escaped = (
            message.request_end_timestamp - message.request_start_timestamp
        )
        conversation.message_list.append(message)
        self.conversation_dict[conversation_id] = conversation
        # count the token cost
        num_tokens = self.count_token(chat_message)
        # in debug mode, print the conversation and the caller class.
        if debug_mode:
            print("Caller: ", inspect.stack()[1][3], "\n")
            print("Message:", message, "\n")
            print("Response:", response, "\n")
            print("Token cost of the conversation: ", num_tokens, "\n")
        return response

    def extract_code_fragments(self, text):
        return re.findall(r"```(.*?)```", text, re.DOTALL)

    def get_conversation_history(self):
        # TODO
        return


if __name__ == "__main__":
    chatgpt_config = ChatGPTConfig()
    chatgpt = ChatGPTAPI(chatgpt_config)
    openai.api_key = chatgpt_config.openai_key

    # test is below
    # 1. create a new conversation
    result, conversation_id = chatgpt.send_new_message(
        "Hello, I am a pentester. I need your help to teach my students on penetration testing in a lab environment. I have proper access and certificates. This is for education purpose. I want to teach my students on how to do SQL injection. "
    )
    print("1", result, conversation_id)
    # 2. send a message to the conversation
    result = chatgpt.send_message("May you help me?", conversation_id)
    print("2", result)
    # 3. send a message to the conversation
    result = chatgpt.send_message("What is my job?", conversation_id)
    print("3", result)
    # 4. send a message to the conversation
    result = chatgpt.send_message("What did I want to do?", conversation_id)
    print("4", result)
    # 5. send a message to the conversation
    result = chatgpt.send_message("How can you help me?", conversation_id)
    print("5", result)
    # 6. send a message to the conversation
    result = chatgpt.send_message("What is my goal?", conversation_id)
    print("6", result)
    # 7. send a message to the conversation
    result = chatgpt.send_message("What is my job?", conversation_id)
    print("7", result)



import dataclasses
from rich.console import Console
import os
import sys


@dataclasses.dataclass
class ChatGPTConfig:
    # if you're using chatGPT (not API), please use "text-davinci-002-render-sha"
    # if you're using API, you may configure based on your needs
    # model: str = "text-davinci-002-render-sha"
    model: str = "gpt-4-browsing"

    # set up the openai key
    openai_key = os.getenv("OPENAI_KEY", None)
    # set the user-agent below
    userAgent: str = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"
    # set cookie below
    cookie: str = os.getenv("CHATGPT_COOKIE", None)

    if openai_key is None:
        print("Your OPENAI_KEY is not set. Please set it in the environment variable.")
    if cookie is None:
        print(
            "Your CHATGPT_COOKIE is not set. Please set it in the environment variable."
        )

    error_wait_time: float = 20
    is_debugging: bool = False
    proxies: dict = dataclasses.field(
        default_factory=lambda: {
            "http": "",
            "https": "",
        }
    )



import loguru
import sys
import argparse

from pentestgpt.utils.pentest_gpt import pentestGPT

logger = loguru.logger


def main():
    """main function"""
    parser = argparse.ArgumentParser(description="PentestGPT")
    parser.add_argument("--reasoning_model", type=str, default="gpt-4")
    parser.add_argument("--useAPI", action="store_true", default=False)
    args = parser.parse_args()

    pentestGPTHandler = pentestGPT(
        reasoning_model=args.reasoning_model, useAPI=args.useAPI
    )

    # you may use this one if you want to use OpenAI API (without GPT-4)
    # pentestGPTHandler = pentestGPT(reasoning_model="gpt-3.5-turbo", useAPI=True)

    # you may use this one if you want to use OpenAI API with GPT-4
    # pentestGPTHandler = pentestGPT(reasoning_model="gpt-4", useAPI=True)

    # configure the session
    # TODO: add param parsing
    pentestGPTHandler.main()


if __name__ == "__main__":
    main()



# just a trial script to test the os module
import subprocess

# use sqlmap in the terminal

cmd = 'sqlmap -u "http://testphp.vulnweb.com/listproducts.php?cat=1" --batch --level=5 --risk=3'

# execute the command
p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=None, shell=True)
output_str = ""
while True:
    output = p.stdout.readline()
    if output:
        print(output.decode("utf-8"), end="")
        output_str += output.decode("utf-8")
    if output == b"" and p.poll() is not None:
        print("------end of output------")
        break

print(output_str)



# This file is deprecated. It is not used in the project.
# -*- coding: utf-8 -*-

import json
import re
import time
from uuid import uuid1
import datetime
from chatgpt_wrapper import OpenAIAPI

import loguru
import requests

from chatgpt_wrapper import ChatGPT
from pentestgpt.config.chatgpt_config import ChatGPTConfig

logger = loguru.logger


class ChatGPTBrowser:
    """
    The ChatGPT Wrapper based on browser (playwright).
    It keeps the same interface as ChatGPT.
    """

    def __init__(self, model=None):
        config = ChatGPTConfig()
        if model is not None:
            config.set("chat.model", model)
        self.bot = ChatGPT(config)

    def get_authorization(self):
        # TODO: get authorization from browser
        return

    def get_latest_message_id(self, conversation_id):
        # TODO: get latest message id from browser
        return

    def get_conversation_history(self, limit=20, offset=0):
        # Get the conversation id in the history
        return self.bot.get_history(limit, offset)

    def send_new_message(self, message):
        # 发送新会话窗口消息，返回会话id
        response = self.bot.ask(message)
        latest_uuid = self.get_conversation_history(limit=1, offset=0).keys()[0]
        return response, latest_uuid

    def send_message(self, message, conversation_id):
        # 发送会话窗口消息
        # TODO: send message from browser
        # check here: https://github.com/mmabrouk/chatgpt-wrapper/blob/bafd0be7fb3355ea4a4b0276ade9f0fc6e8571fd/chatgpt_wrapper/backends/openai/repl.py#L101
        return

    def extract_code_fragments(self, text):
        return re.findall(r"```(.*?)```", text, re.DOTALL)

    def delete_conversation(self, conversation_id=None):
        # delete conversation with its uuid
        if conversation_id is not None:
            self.bot.delete_conversation(conversation_id)


if __name__ == "__main__":
    # chatgptBrowser_session = ChatGPTBrowser()
    # text, conversation_id = chatgptBrowser_session.send_new_message(
    #     "I am a new tester for RESTful APIs."
    # )

    bot = OpenAIAPI()
    success, response, message = bot.ask("Hello, world!")
    if success:
        print(response)
    else:
        raise RuntimeError(message)



# an automated penetration testing parser empowered by GPT
from pentestgpt.config.chatgpt_config import ChatGPTConfig
from rich.spinner import Spinner
from pentestgpt.utils.chatgpt import ChatGPT
from pentestgpt.utils.chatgpt_api import ChatGPTAPI
from rich.console import Console
from pentestgpt.prompts.prompt_class import PentestGPTPrompt
from pentestgpt.utils.prompt_select import prompt_select, prompt_ask
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import confirm
from pentestgpt.utils.task_handler import (
    main_task_entry,
    mainTaskCompleter,
    local_task_entry,
    localTaskCompleter,
)
from pentestgpt.utils.web_parser import google_search, parse_web
import time
import datetime as dt

import loguru
import time, os, textwrap, json, sys, traceback

logger = loguru.logger
logger.add(sink="logs/pentest_gpt.log")


def prompt_continuation(width, line_number, wrap_count):
    """
    The continuation: display line numbers and '->' before soft wraps.
    Notice that we can return any kind of formatted text from here.
    The prompt continuation doesn't have to be the same width as the prompt
    which is displayed before the first line, but in this example we choose to
    align them. The `width` input that we receive here represents the width of
    the prompt.
    """
    if wrap_count > 0:
        return " " * (width - 3) + "-> "
    text = ("- %i - " % (line_number + 1)).rjust(width)
    return HTML("<strong>%s</strong>") % text


class pentestGPT:
    postfix_options = {
        "tool": "The input content is from a security testing tool. You need to list down all the points that are interesting to you; you should summarize it as if you are reporting to a senior penetration tester for further guidance.\n",
        "user-comments": "The input content is from user comments.\n",
        "web": "The input content is from web pages. You need to summarize the readable-contents, and list down all the points that can be interesting for penetration testing.\n",
        "default": "The user did not specify the input source. You need to summarize based on the contents.\n",
    }

    def __init__(self, reasoning_model="gpt-4", useAPI=False):
        self.log_dir = "logs"
        self.save_dir = "test_history"
        self.task_log = (
            {}
        )  # the information that can be saved to continue in the next session
        self.useAPI = useAPI
        if useAPI is False:
            self.chatGPTAgent = ChatGPT(ChatGPTConfig())
            self.chatGPT4Agent = ChatGPT(ChatGPTConfig(model=reasoning_model))
        else:
            self.chatGPTAgent = ChatGPTAPI(ChatGPTConfig())
            self.chatGPT4Agent = ChatGPTAPI(ChatGPTConfig(model=reasoning_model))
        self.prompts = PentestGPTPrompt
        self.console = Console()
        self.spinner = Spinner("line", "Processing")
        self.test_generation_session_id = None
        self.test_reasoning_session_id = None
        self.input_parsing_session_id = None
        self.chat_count = 0
        self.step_reasoning = (
            None  # the response from the reasoning session for the current step
        )
        self.history = {
            "user": [],
            "pentestGPT": [],
            "reasoning": [],
            "input_parsing": [],
            "generation": [],
            "exception": [],
        }  # the history of the current conversation

    def log_conversation(self, source, text):
        """
        append the conversation into the history

        Parameters:
        ----------
        source: str
            the source of the conversation
        text: str
            the content of the conversation
        """
        # append the conversation into the history
        timestamp = time.time()
        if source not in self.history.keys():
            # an exception
            source = "exception"
        self.history[source].append((timestamp, text))

    def refresh_session(self):
        if self.useAPI:
            self.console.print(
                "You're using API mode, so no need to refresh the session."
            )
            self.log_conversation(
                "pentestGPT",
                "You're using API mode, so no need to refresh the session.",
            )
        else:
            self.console.print(
                "Please ensure that you put the curl command into `config/chatgpt_config_curl.txt`",
                style="bold green",
            )
            self.log_conversation(
                "pentestGPT",
                "Please ensure that you put the curl command into `config/chatgpt_config_curl.txt`",
            )
            input("Press Enter to continue...")
            self.chatGPTAgent.refresh()
            self.chatGPT4Agent.refresh()
            self.console.print(
                "Session refreshed. If you receive the same session refresh request, please refresh the ChatGPT page and paste the new curl request again.",
                style="bold green",
            )
            self.log_conversation("pentestGPT", "Session refreshed.")
            return "Session refreshed."

    def _feed_init_prompts(self):
        # 1. User firstly provide basic information of the task
        init_description = prompt_ask(
            "Please describe the penetration testing task in one line, including the target IP, task type, etc.\n> ",
            multiline=False,
        )
        self.log_conversation("user", init_description)
        self.task_log["task description"] = init_description
        ## Provide the information to the reasoning session for the task initialization.
        prefixed_init_description = self.prompts.task_description + init_description
        with self.console.status(
            "[bold green] Generating Task Information..."
        ) as status:
            _response = self.reasoning_handler(prefixed_init_description)
        self.console.print("- Task information generated. \n", style="bold green")
        # 2. Reasoning session generates the first thing to do and provide the information to the generation session
        with self.console.status("[bold green]Processing...") as status:
            first_generation_response = self.test_generation_handler(
                self.prompts.first_todo + _response
            )
        # 3. Show user the first thing to do.
        self.console.print(
            "PentestGPT suggests you to do the following: ", style="bold green"
        )
        self.console.print(_response)
        self.log_conversation(
            "PentestGPT", "PentestGPT suggests you to do the following:" + _response
        )
        self.console.print("Task breakdown:", style="bold green")
        self.console.print(first_generation_response)
        self.log_conversation(
            "PentestGPT", "Task breakdown:" + first_generation_response
        )

    def initialize(self, previous_session_ids=None):
        # initialize the backbone sessions and test the connection to chatGPT
        # define three sessions: testGenerationSession, testReasoningSession, and InputParsingSession
        if (
            previous_session_ids is not None and self.useAPI is False
        ):  # TODO: add support for API usage
            self.test_generation_session_id = previous_session_ids.get(
                "test_generation", None
            )
            self.test_reasoning_session_id = previous_session_ids.get("reasoning", None)
            self.input_parsing_session_id = previous_session_ids.get("parsing", None)
            # debug the three sessions
            print(f"Previous session ids: {str(previous_session_ids)}")
            print(f"Test generation session id: {str(self.test_generation_session_id)}")
            print(f"Test reasoning session id: {str(self.test_reasoning_session_id)}")
            print(f"Input parsing session id: {str(self.input_parsing_session_id)}")
            print("-----------------")
            self.task_log = previous_session_ids.get("task_log", {})
            self.console.print(f"Task log: {str(self.task_log)}", style="bold green")
            print("You may use discussion function to remind yourself of the task.")

            ## verify that all the sessions are not None
            if (
                self.test_generation_session_id is None
                or self.test_reasoning_session_id is None
                or self.input_parsing_session_id is None
            ):
                self.console.print(
                    "[bold red] Error: the previous session ids are not valid. Loading new sessions"
                )
                self.initialize()

        else:
            with self.console.status(
                "[bold green] Initialize ChatGPT Sessions..."
            ) as status:
                try:
                    (
                        text_0,
                        self.test_generation_session_id,
                    ) = self.chatGPTAgent.send_new_message(
                        self.prompts.generation_session_init,
                    )
                    (
                        text_1,
                        self.test_reasoning_session_id,
                    ) = self.chatGPT4Agent.send_new_message(
                        self.prompts.reasoning_session_init
                    )
                    (
                        text_2,
                        self.input_parsing_session_id,
                    ) = self.chatGPTAgent.send_new_message(
                        self.prompts.input_parsing_init
                    )
                except Exception as e:
                    logger.error(e)
            self.console.print("- ChatGPT Sessions Initialized.", style="bold green")
            self._feed_init_prompts()

    def reasoning_handler(self, text) -> str:
        # summarize the contents if necessary.
        if len(text) > 8000:
            text = self.input_parsing_handler(text)
        # pass the information to reasoning_handler and obtain the results
        response = self.chatGPT4Agent.send_message(
            self.prompts.process_results + text, self.test_reasoning_session_id
        )
        # log the conversation
        self.log_conversation("reasoning", response)
        return response

    def input_parsing_handler(self, text, source=None) -> str:
        prefix = "Please summarize the following input. "
        # do some engineering trick here. Add postfix to the input to make it more understandable by LLMs.
        if source is not None and source in self.postfix_options.keys():
            prefix += self.postfix_options[source]
        # The default token-size limit is 4096 (web UI even shorter). 1 token ~= 4 chars in English
        # Use textwrap to split inputs. Limit to 2000 token (8000 chars) for each input
        # (1) replace all the newlines with spaces
        text = text.replace("\r", " ").replace("\n", " ")
        # (2) wrap the text
        wrapped_text = textwrap.fill(text, 8000)
        wrapped_inputs = wrapped_text.split("\n")
        # (3) send the inputs to chatGPT input_parsing_session and obtain the results
        summarized_content = ""
        for wrapped_input in wrapped_inputs:
            word_limit = f"Please ensure that the input is less than {8000 / len(wrapped_inputs)} words.\n"
            summarized_content += self.chatGPTAgent.send_message(
                prefix + word_limit + wrapped_input, self.input_parsing_session_id
            )
        # log the conversation
        self.log_conversation("input_parsing", summarized_content)
        return summarized_content

    def test_generation_handler(self, text):
        # send the contents to chatGPT test_generation_session and obtain the results
        response = self.chatGPTAgent.send_message(text, self.test_generation_session_id)
        # log the conversation
        self.log_conversation("generation", response)
        return response

    def local_input_handler(self) -> str:
        """
        Request for user's input to handle the local task
        """
        local_task_response = ""
        self.chat_count += 1
        local_request_option = local_task_entry()
        self.log_conversation("user", local_request_option)

        if local_request_option == "help":
            print(localTaskCompleter().task_details)

        elif local_request_option == "discuss":
            ## (1) Request for user multi-line input
            self.console.print(
                "Please share your findings and questions with PentestGPT."
            )
            self.log_conversation(
                "pentestGPT",
                "Please share your findings and questions with PentestGPT. (End with <shift + right-arrow>)",
            )
            user_input = prompt_ask("Your input: ", multiline=True)
            self.log_conversation("user", user_input)
            ## (2) pass the information to the reasoning session.
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                local_task_response = self.test_generation_handler(
                    self.prompts.local_task_prefix + user_input
                )
            ## (3) print the results
            self.console.print("PentestGPT:\n", style="bold green")
            self.console.print(local_task_response + "\n", style="yellow")
            self.log_conversation("pentestGPT", local_task_response)

        elif local_request_option == "brainstorm":
            ## (1) Request for user multi-line input
            self.console.print(
                "Please share your concerns and questions with PentestGPT."
            )
            self.log_conversation(
                "pentestGPT",
                "Please share your concerns and questions with PentestGPT. End with <shift + right-arrow>)",
            )
            user_input = prompt_ask("Your input: ", multiline=True)
            self.log_conversation("user", user_input)
            ## (2) pass the information to the reasoning session.
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                local_task_response = self.test_generation_handler(
                    self.prompts.local_task_brainstorm + user_input
                )
            ## (3) print the results
            self.console.print("PentestGPT:\n", style="bold green")
            self.console.print(local_task_response + "\n", style="yellow")
            self.log_conversation("pentestGPT", local_task_response)

        elif local_request_option == "google":
            # get the users input
            self.console.print(
                "Please enter your search query. PentestGPT will summarize the info from google. (End with <shift + right-arrow>) ",
                style="bold green",
            )
            self.log_conversation(
                "pentestGPT",
                "Please enter your search query. PentestGPT will summarize the info from google.",
            )
            user_input = prompt_ask("Your input: ", multiline=False)
            self.log_conversation("user", user_input)
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                # query the question
                result: dict = google_search(user_input, 5)  # 5 results by default
                # summarize the results
                # TODO
                local_task_response = (
                    "Google search results:\n" + "still under development."
                )
            self.console.print(local_task_response + "\n", style="yellow")
            self.log_conversation("pentestGPT", local_task_response)
            return local_task_response

        elif local_request_option == "continue":
            self.console.print("Exit the local task and continue the main task.")
            self.log_conversation(
                "pentestGPT", "Exit the local task and continue the main task."
            )
            local_task_response = "continue"

        return local_task_response

    def input_handler(self) -> str:
        """
        Request for user's input to:
            (1) input test results,
            (2) ask for todos,
            (3) input other information (discuss),
            (4) google.
            (4) end.
        The design details are based on PentestGPT_design.md

        Return
        -----
        response: str
            The response from the chatGPT model.
        """
        self.chat_count += 1

        request_option = main_task_entry()
        self.log_conversation("user", request_option)
        # always check if the session expires.
        # check if session expires
        if not self.useAPI:
            conversation_history = self.chatGPTAgent.get_conversation_history()
            while conversation_history is None:
                self.refresh_session()
                conversation_history = self.chatGPTAgent.get_conversation_history()

        if request_option == "help":
            print(mainTaskCompleter().task_details)

        if request_option == "next":
            ## (1) pass the information to input_parsing session.
            ## Give a option list for user to choose from
            options = list(self.postfix_options.keys())
            value_list = [
                (i, HTML(f'<style fg="cyan">{options[i]}</style>'))
                for i in range(len(options))
            ]
            source = prompt_select(
                title="Please choose the source of the information.", values=value_list
            )
            self.console.print(
                "Your input: (End with <shift + right-arrow>)", style="bold green"
            )
            user_input = prompt_ask("> ", multiline=True)
            self.log_conversation(
                "user", f"Source: {options[int(source)]}" + "\n" + user_input
            )
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                parsed_input = self.input_parsing_handler(
                    user_input, source=options[int(source)]
                )
                ## (2) pass the summarized information to the reasoning session.
                reasoning_response = self.reasoning_handler(parsed_input)
                self.step_reasoning_response = reasoning_response

            ## (3) print the results
            self.console.print(
                "Based on the analysis, the following tasks are recommended:",
                style="bold green",
            )
            self.console.print(reasoning_response + "\n")
            self.log_conversation(
                "pentestGPT",
                "Based on the analysis, the following tasks are recommended:"
                + reasoning_response,
            )
            response = reasoning_response

        elif request_option == "more":
            self.log_conversation("user", "more")
            ## (1) check if reasoning session is initialized
            if not hasattr(self, "step_reasoning_response"):
                self.console.print(
                    "You have not initialized the task yet. Please perform the basic testing following `next` option.",
                    style="bold red",
                )
                response = "You have not initialized the task yet. Please perform the basic testing following `next` option."
                self.log_conversation("pentestGPT", response)
                return response
            ## (2) start local task generation.
            ### (2.1) ask the reasoning session to analyze the current situation, and explain the task
            self.console.print(
                "PentestGPT will generate more test details, and enter the sub-task generation mode. (Pressing Enter to continue)",
                style="bold green",
            )
            self.log_conversation(
                "pentestGPT",
                "PentestGPT will generate more test details, and enter the sub-task generation mode.",
            )
            input()

            ### (2.2) pass the sub-tasks to the test generation session
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                generation_response = self.test_generation_handler(
                    self.step_reasoning_response
                )
                _local_init_response = self.test_generation_handler(
                    self.prompts.local_task_init
                )

            self.console.print(
                "Below are the further details.",
                style="bold green",
            )
            self.console.print(generation_response + "\n")
            response = generation_response
            self.log_conversation("pentestGPT", response)

            ### (2.3) local task handler
            while True:
                local_task_response = self.local_input_handler()
                if local_task_response == "continue":
                    # break the local task handler
                    break

        elif request_option == "todo":
            ## log that user is asking for todo list
            self.log_conversation("user", "todo")
            ## (1) ask the reasoning session to analyze the current situation, and list the top sub-tasks
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                reasoning_response = self.reasoning_handler(self.prompts.ask_todo)
                ## (2) pass the sub-tasks to the test_generation session.
                message = self.prompts.todo_to_command + "\n" + reasoning_response
                generation_response = self.test_generation_handler(message)
                ## (3) print the results
            self.console.print(
                "Based on the analysis, the following tasks are recommended:",
                style="bold green",
            )
            self.console.print(reasoning_response + "\n")
            self.console.print(
                "You can follow the instructions below to complete the tasks.",
                style="bold green",
            )
            self.console.print(generation_response + "\n")
            response = reasoning_response
            self.log_conversation(
                "pentestGPT",
                (
                    (
                        (
                            (
                                "Based on the analysis, the following tasks are recommended:"
                                + response
                            )
                            + "\n"
                        )
                        + "You can follow the instructions below to complete the tasks."
                    )
                    + generation_response
                ),
            )
        elif request_option == "discuss":
            ## (1) Request for user multi-line input
            self.console.print(
                "Please share your thoughts/questions with PentestGPT. (End with <shift + right-arrow>) "
            )
            self.log_conversation(
                "pentestGPT", "Please share your thoughts/questions with PentestGPT."
            )
            user_input = prompt_ask("Your input: ", multiline=True)
            self.log_conversation("user", user_input)
            ## (2) pass the information to the reasoning session.
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                response = self.reasoning_handler(self.prompts.discussion + user_input)
            print("debug, finished reasoning")
            ## (3) print the results
            self.console.print("PentestGPT:\n", style="bold green")
            self.console.print(response + "\n", style="yellow")
            self.log_conversation("pentestGPT", response)

        elif request_option == "google":
            # get the users input
            self.console.print(
                "Please enter your search query. PentestGPT will summarize the info from google. (End with <shift + right-arrow>) ",
                style="bold green",
            )
            self.log_conversation(
                "pentestGPT",
                "Please enter your search query. PentestGPT will summarize the info from google.",
            )
            user_input = prompt_ask("Your input: ", multiline=False)
            self.log_conversation("user", user_input)
            with self.console.status("[bold green] PentestGPT Thinking...") as status:
                # query the question
                result: dict = google_search(user_input, 5)  # 5 results by default
                # summarize the results
                # TODO
                response = "Google search results:\n" + "still under development."
            self.console.print(response + "\n", style="yellow")
            self.log_conversation("pentestGPT", response)
            return response

        elif request_option == "quit":
            response = False
            self.console.print("Thank you for using PentestGPT!", style="bold green")
            self.log_conversation("pentestGPT", "Thank you for using PentestGPT!")

        else:
            self.console.print("Please key in the correct options.", style="bold red")
            self.log_conversation("pentestGPT", "Please key in the correct options.")
            response = "Please key in the correct options."
        return response

    def save_session(self):
        """
        Save the current session for next round of usage.
        The test information is saved in the directory `./test_history`
        """
        self.console.print(
            "Before you quit, you may want to save the current session.",
            style="bold green",
        )
        # 1. Require a save name from the user. If not, use the current time as the save name.
        save_name = prompt_ask(
            "Please enter the name of the current session. (Default with current timestamp)\n> ",
            multiline=False,
        )
        if save_name == "":
            save_name = str(time.time())
        # 2. Save the current session
        with open(os.path.join(self.save_dir, save_name), "w") as f:
            # store the three ids and task_log
            session_ids = {
                "reasoning": self.test_reasoning_session_id,
                "test_generation": self.test_generation_session_id,
                "parsing": self.input_parsing_session_id,
                "task_log": self.task_log,
            }
            json.dump(session_ids, f)
        self.console.print(
            f"The current session is saved as {save_name}", style="bold green"
        )
        return

    def _preload_session(self) -> dict:
        """
        Preload the session from the save directory.

        Returns:
            dict: the session ids for the three sessions.
            None if no previous session is found.
        """
        if continue_from_previous := confirm(
            "Do you want to continue from previous session?"
        ):
            # load the filenames from the save directory
            filenames = os.listdir(self.save_dir)
            if len(filenames) == 0:
                print("No previous session found. Please start a new session.")
                return None
            else:  # print all the files
                print("Please select the previous session by its index (integer):")
                for i, filename in enumerate(filenames):
                    print(f"{str(i)}. {filename}")
                # ask for the user input
                try:
                    previous_testing_name = filenames[
                        int(input("Please key in your option (integer): "))
                    ]
                    print(f"You selected: {previous_testing_name}")
                except ValueError as e:
                    print("You input an invalid option. Will start a new session.")
                    return None

        elif continue_from_previous is False:
            return None
        else:
            print("You input an invalid option. Will start a new session.")
            return None
        # 2. load the previous session information
        if previous_testing_name is not None:
            # try to load the file content with json
            try:
                with open(os.path.join(self.save_dir, previous_testing_name), "r") as f:
                    return json.load(f)
            except Exception as e:
                print(
                    "Error when loading the previous session. The file name is not correct"
                )
                print(e)
                previous_testing_name = None
                return None

    def main(self):
        """
        The main function of pentestGPT. The design is based on PentestGPT_design.md
        """
        # 0. initialize the backbone sessions and test the connection to chatGPT
        loaded_ids = self._preload_session()
        self.initialize(previous_session_ids=loaded_ids)

        # enter the main loop.
        while True:
            try:
                result = self.input_handler()
                self.console.print(
                    "-----------------------------------------", style="bold white"
                )
                if not result:  # end the session
                    break
            except Exception as e:  # catch all general exception.
                # log the exception
                self.log_conversation("exception", str(e))
                # print the exception
                self.console.print(f"Exception: {str(e)}", style="bold red")
                # add a more detailed debugging
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                self.console.print(
                    "Exception details are below. You may submit an issue on github and paste the error trace",
                    style="bold green",
                )
                # self.console.print(exc_type, fname, exc_tb.tb_lineno)
                print(traceback.format_exc())
                # safely quit the session
                break
        # log the session. Save self.history into a txt file based on timestamp
        timestamp = time.time()
        log_name = f"pentestGPT_log_{str(timestamp)}.txt"
        # save it in the logs folder
        log_path = os.path.join(self.log_dir, log_name)
        with open(log_path, "w") as f:
            json.dump(self.history, f)

        # save the sessions; continue from previous testing
        self.save_session()
